/* openpgp-oids.c - OID helper for OpenPGP
 * Copyright (C) 2024 Free Software Foundation, Inc.
 *
 * This file is part of GnuPG.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either
 *
 *   - the GNU Lesser General Public License as published by the Free
 *     Software Foundation; either version 3 of the License, or (at
 *     your option) any later version.
 *
 * or
 *
 *   - the GNU General Public License as published by the Free
 *     Software Foundation; either version 2 of the License, or (at
 *     your option) any later version.
 *
 * or both in parallel, as here.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 */

#include "sm2.h"

const unsigned char zin_default[] = {
  0x00, 0x80, // id length
  0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x31, 0x32, 0x33, 0x34,
  0x35, 0x36, 0x37, 0x38, // default id: 1234567812345678
  0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc, // sm2 a
  0x28, 0xe9, 0xfa, 0x9e, 0x9d, 0x9f, 0x5e, 0x34, 0x4d, 0x5a, 0x9e, 0x4b,
  0xcf, 0x65, 0x09, 0xa7, 0xf3, 0x97, 0x89, 0xf5, 0x15, 0xab, 0x8f, 0x92,
  0xdd, 0xbc, 0xbd, 0x41, 0x4d, 0x94, 0x0e, 0x93, // sm2 b
  0x32, 0xc4, 0xae, 0x2c, 0x1f, 0x19, 0x81, 0x19, 0x5f, 0x99, 0x04, 0x46,
  0x6a, 0x39, 0xc9, 0x94, 0x8f, 0xe3, 0x0b, 0xbf, 0xf2, 0x66, 0x0b, 0xe1,
  0x71, 0x5a, 0x45, 0x89, 0x33, 0x4c, 0x74, 0xc7, // sm2 x
  0xbc, 0x37, 0x36, 0xa2, 0xf4, 0xf6, 0x77, 0x9c, 0x59, 0xbd, 0xce, 0xe3,
  0x6b, 0x69, 0x21, 0x53, 0xd0, 0xa9, 0x87, 0x7c, 0xc6, 0x2a, 0x47, 0x40,
  0x02, 0xdf, 0x32, 0xe5, 0x21, 0x39, 0xf0, 0xa0 // sm2 y
};

static gpg_error_t
calculate_za (gcry_mpi_t pkey, unsigned char *result)
{
  gpg_error_t err;
  gcry_md_hd_t md;
  unsigned char q[65];
  size_t qlen;

  err = gcry_mpi_print (GCRYMPI_FMT_STD, q, sizeof(q), &qlen, pkey);
  if (err)
    return err;

  err = gcry_md_open (&md, GCRY_MD_SM3_PGP, 0);
  if (err)
    return err;

  gcry_md_write (md, zin_default, sizeof(zin_default));
  gcry_md_write (md, q + 1, qlen - 1);
  gcry_md_final (md);
  memcpy (result, gcry_md_read (md, GCRY_MD_SM3_PGP), 32);
  gcry_md_close (md);

  return 0;
}

gpg_error_t
update_sm2_hash_md (gcry_md_hd_t md, gcry_mpi_t pkey)
{
  gpg_error_t err;
  unsigned char za[32];
  unsigned char msg[32];

  if (!md || !pkey)
    return gpg_error (GPG_ERR_INV_VALUE);

  err = calculate_za(pkey, za);
  if (err)
    return err;

  memcpy(msg, gcry_md_read(md, GCRY_MD_SM3_PGP), 32);
  gcry_md_reset(md);
  gcry_md_write(md, za, sizeof(za));
  gcry_md_write(md, msg, sizeof(msg));
  gcry_md_final(md);

  return 0; 
}

gpg_error_t 
update_sm2_hash_value (unsigned char *msg, size_t msglen, gcry_sexp_t s_pkey, unsigned char *result)
{
  gpg_error_t err;
  gcry_sexp_t q_sexp;
  gcry_mpi_t q_mpi;
  gcry_md_hd_t md;
  unsigned char za[32];

  if(!msg || !s_pkey || !result)
    return gpg_error (GPG_ERR_INV_VALUE);

  q_sexp = gcry_sexp_find_token(s_pkey, "q", 0);
  if (!q_sexp)
    return gpg_error (GPG_ERR_NO_PUBKEY);

  q_mpi = gcry_sexp_nth_mpi (q_sexp, 1, GCRYMPI_FMT_USG);
  gcry_sexp_release (q_sexp);
  if (!q_mpi)
    return gpg_error (GPG_ERR_NO_PUBKEY);

  err = calculate_za (q_mpi, za);
  gcry_mpi_release (q_mpi);
  if (err)
    return err;

  err = gcry_md_open (&md, GCRY_MD_SM3_PGP, 0);
  if (err)
    return err;

  gcry_md_write (md, za, sizeof(za));
  gcry_md_write (md, msg, msglen);
  gcry_md_final (md);
  memcpy (result, gcry_md_read (md, GCRY_MD_SM3_PGP), 32);
  gcry_md_close (md);

  return 0;
}

int pkey_is_sm2 (gcry_sexp_t pkey)
{
  const char *curve;

  if (!pkey)
    return 0;

  curve = gcry_pk_get_curve (pkey, 0, NULL);
  if (curve && !strcmp (curve, "sm2p256v1"))
    return 1;

  return 0;
}
